/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wAdjoint.h"
#include "wProblem.h"
#include "wFluid.h"
#include "wFreestream.h"
#include "wWake.h"
#include "wKutta.h"
#include "wBody.h"
#include "wNewton.h"
#include "wPotentialResidual.h"
#include "wFreestreamResidual.h"
#include "wWakeResidual.h"
#include "wKuttaResidual.h"
#include "wLoadFunctional.h"

#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wLinearSolver.h"
#include "wGmres.h"
#include "wMshDeform.h"
#include "wResults.h"
#include "wMshExport.h"

#include <Eigen/Sparse>

#include <tbb/global_control.h>
#include <tbb/parallel_for_each.h>
#include <tbb/spin_mutex.h>

#include <iomanip>
#include <deque>

using namespace tbox;
using namespace dart;

/**
 * @brief Initialize the adjoint solver
 */
Adjoint::Adjoint(std::shared_ptr<Newton> _sol, std::shared_ptr<tbox::MshDeform> _morph) : sol(_sol), morph(_morph)
{
    // Default parameters
    nthreads = sol->nthreads;
    verbose = sol->verbose;

    // Linear solvers (if GMRES, use the same, but with a thighter tolerance)
    if (sol->linsol->type() == LSolType::GMRES_ILUT)
    {
        std::shared_ptr<Gmres> gmres = std::make_shared<Gmres>(*dynamic_cast<Gmres *>(sol->linsol.get()));
        gmres->setTolerance(1e-8);
        alinsol = gmres;
    }
    else
        alinsol = sol->linsol;
    if (morph->linsol->type() == LSolType::GMRES_ILUT)
    {
        std::shared_ptr<Gmres> gmres = std::make_shared<Gmres>(*dynamic_cast<Gmres *>(morph->linsol.get()));
        gmres->setTolerance(1e-12);
        mlinsol = gmres;
    }
    else
        mlinsol = morph->linsol;

    // Update element gradients
    sol->pbl->initGradElems();

    // Setup adjoint variables and gradients
    lamClFlo.resize(sol->pbl->msh->nodes.size(), 0.);
    lamCdFlo.resize(sol->pbl->msh->nodes.size(), 0.);
    dClMsh.resize(sol->pbl->msh->nodes.size(), Eigen::Vector3d::Zero());
    dCdMsh.resize(sol->pbl->msh->nodes.size(), Eigen::Vector3d::Zero());
    dClAoa = 0;
    dCdAoa = 0;

    // Display solver parameters
    std::cout << "--- Adjoint solver ---\n"
              << "Aero inner solver: " << *alinsol
              << "Mesh inner solver: " << *mlinsol
              << "Number of threads: " << nthreads << "\n"
              << std::endl;
}

/**
 * @brief Run the adjoint method
 *
 * Solve the adjoint steady transonic Full Potential Equation
 * and compute lift and drag sensitivites with respect to angle of attack and
 * mesh coordinates for aerodynamic optimization (single-discipline)
 */
void Adjoint::run()
{
    // Init
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);

    // Display current freestream conditions
    if (verbose > 0)
        std::cout << std::fixed << std::setprecision(2)
                  << "Computing gradients for Mach " << sol->pbl->M_inf << ", "
                  << sol->pbl->alpha * 180 / 3.14159 << "deg AoA, "
                  << sol->pbl->beta * 180 / 3.14159 << "deg AoS"
                  << std::endl;

    // Compute partial gradients of flow residuals and solve flow adjoint
    tms["0-AdjFlo"].start();
    this->linearizeFlow();                                                            // dRu/dU, dRu/dA, dRu/dX
    std::vector<std::vector<double>> dCf_U = this->computeGradientCoefficientsFlow(); // {dCl, dCd}/dU
    lamClFlo = this->solveFlow(dCf_U[0]);                                             // lambdaClU = dRu/dU^-T * dCl/dU
    lamCdFlo = this->solveFlow(dCf_U[1]);                                             // lambdaCdU = dRu/dU^-T * dCd/dU
    tms["0-AdjFlo"].stop();

    // Compute gradients with respect to AoA
    tms["1-GradAoA"].start();
    // partials
    std::vector<double> dCf_A = this->computeGradientCoefficientsAoa(); // {dCl, dCd}/dA
    // totals
    dClAoa = dCf_A[0] - this->computeJacVecFlowAoa(lamClFlo); // dCl/dA - dRu/dA^T * lambdaClU
    dCdAoa = dCf_A[1] - this->computeJacVecFlowAoa(lamCdFlo); // dCd/dA - dRu/dA^T * lambdaCdU
    tms["1-GradAoA"].stop();

    // Compute gradients with respect to mesh coordinates
    tms["2-GradMsh"].start();
    // compute mesh Jacobian
    this->linearizeMesh(); // dRx/dX
    // partials
    std::vector<std::vector<double>> dCf_X = this->computeGradientCoefficientsMesh(); // {dCl, dCd}/dX
    std::vector<double> dCl_XU = this->computeJacVecFlowMesh(lamClFlo);
    std::vector<double> dCd_XU = this->computeJacVecFlowMesh(lamCdFlo);
    for (size_t i = 0; i < sol->pbl->nDim * sol->pbl->msh->nodes.size(); ++i)
    {
        dCf_X[0][i] -= dCl_XU[i]; // dCl/dX - dRu/dX^T * lambdaClU
        dCf_X[1][i] -= dCd_XU[i]; // dCd/dX - dRu/dX^T * lambdaCdU
    }
    // compute mesh adjoint (totals)
    std::vector<double> lamClMsh = this->solveMesh(dCf_X[0]); // lambdaClX = dRx/dX^-T * (dCl/dX - dRu/dX^T * lambdaClU)
    std::vector<double> lamCdMsh = this->solveMesh(dCf_X[1]); // lambdaCdX = dRx/dX^-T * (dCd/dX - dRu/dX^T * lambdaCdU)
    // store solution
    for (auto n : sol->pbl->msh->nodes)
    {
        for (int m = 0; m < sol->pbl->nDim; m++)
        {
            dClMsh[n->row](m) = lamClMsh[sol->pbl->nDim * (n->row) + m];
            dCdMsh[n->row](m) = lamCdMsh[sol->pbl->nDim * (n->row) + m];
        }
    }
    tms["2-GradMsh"].stop();

    // Display
    // flow adjoint
    std::cout << std::setw(15) << std::left << "Flow adjoint"
              << std::setw(15) << std::right << "Res[lambdaCl]"
              << std::setw(15) << std::right << "Res[lambdaCd]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << ""
              << std::setw(15) << std::right << log10((dRu_U.transpose() * Eigen::Map<Eigen::VectorXd>(lamClFlo.data(), lamClFlo.size()) - Eigen::Map<Eigen::VectorXd>(dCf_U[0].data(), dCf_U[0].size())).norm())
              << std::setw(15) << std::right << log10((dRu_U.transpose() * Eigen::Map<Eigen::VectorXd>(lamCdFlo.data(), lamCdFlo.size()) - Eigen::Map<Eigen::VectorXd>(dCf_U[1].data(), dCf_U[1].size())).norm()) << std::endl;
    // aoa gradients
    std::cout << std::setw(15) << std::left << "AoA gradients"
              << std::setw(15) << std::right << "dCl_dAoA"
              << std::setw(15) << std::right << "dCd_dAoA" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << ""
              << std::setw(15) << std::right << dClAoa
              << std::setw(15) << std::right << dCdAoa << std::endl;
    // mesh adjoint
    std::cout << std::setw(15) << std::left << "Mesh adjoint"
              << std::setw(15) << std::right << "Res[lambdaCl]"
              << std::setw(15) << std::right << "Res[lambdaCd]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << ""
              << std::setw(15) << std::right << log10((dRx_X.transpose() * Eigen::Map<Eigen::VectorXd>(lamClMsh.data(), lamClMsh.size()) - Eigen::Map<Eigen::VectorXd>(dCf_X[0].data(), dCf_X[0].size())).norm())
              << std::setw(15) << std::right << log10((dRx_X.transpose() * Eigen::Map<Eigen::VectorXd>(lamCdMsh.data(), lamCdMsh.size()) - Eigen::Map<Eigen::VectorXd>(dCf_X[1].data(), dCf_X[1].size())).norm()) << std::endl;
    // mesh gradients
    std::cout << std::setw(15) << std::left << "Mesh gradients"
              << std::setw(15) << std::right << "Norm[dCl_dX]"
              << std::setw(15) << std::right << "Norm[dCd_dX]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << ""
              << std::setw(15) << std::right << Eigen::Map<Eigen::VectorXd>(lamClMsh.data(), lamClMsh.size()).norm()
              << std::setw(15) << std::right << Eigen::Map<Eigen::VectorXd>(lamCdMsh.data(), lamCdMsh.size()).norm() << std::endl;

    // Display timers
    if (verbose > 2)
        std::cout << "Adjoint solver CPU" << std::endl
                  << tms;
    std::cout << std::endl;
}

/**
 * @brief Compute the mesh Jacobian
 *
 * J = dRx/dX (x is the vector of mesh coordinates)
 */
void Adjoint::linearizeMesh()
{
    dRx_X = Eigen::SparseMatrix<double, Eigen::RowMajor>(sol->pbl->nDim * sol->pbl->msh->nodes.size(), sol->pbl->nDim * sol->pbl->msh->nodes.size());
    morph->build(dRx_X); // J = dRx/dX
}

/**
 * @brief Solve for the mesh residuals given a gradient (wrt to mesh coordinates)
 *
 * solve for dRx: dRx_dX^T * dRx = dX
 */
std::vector<double> Adjoint::solveMesh(std::vector<double> const &dX)
{
    // Create maps
    Eigen::VectorXd dR(sol->pbl->nDim * sol->pbl->msh->nodes.size());
    Eigen::Map<Eigen::VectorXd> dR_(dR.data(), dR.size());
    Eigen::VectorXd dX_ = Eigen::Map<const Eigen::VectorXd>(dX.data(), dX.size());
    // Solve
    mlinsol->compute(dRx_X.transpose(), Eigen::Map<Eigen::VectorXd>(dX_.data(), dX_.size()), dR_);
    return std::vector<double>(dR.data(), dR.data() + dR.size());
}

/**
 * @brief Compute the matrix-vector product of the mesh Jacobian with the mesh residual
 *
 * dX = dRx/dX^T * dRx
 * Note: only for multi-disciplinary, not tested by run()
 */
std::vector<double> Adjoint::computeJacVecMesh(std::vector<double> const &dR)
{
    Eigen::VectorXd JdR = dRx_X.transpose() * Eigen::Map<const Eigen::VectorXd>(dR.data(), dR.size());
    return std::vector<double>(JdR.data(), JdR.data() + JdR.size());
}

/**
 * @brief Compute the partial gradients of the flow residuals
 *
 * dRu_U = dRu/dU (u is the vector of flow variables, i.e. the potential "phi")
 * dRu_A = dRu/dA (a is the aoa)
 * dRu_X = dRu/dX (x is the vector of mesh coordinates)
 */
void Adjoint::linearizeFlow()
{
    // Partials wrt flow (Jacobian)
    dRu_U = Eigen::SparseMatrix<double, Eigen::RowMajor>(sol->pbl->msh->nodes.size(), sol->pbl->msh->nodes.size());
    sol->buildJac(dRu_U);
    // Partials wrt AoA
    dRu_A = Eigen::VectorXd::Zero(sol->pbl->msh->nodes.size());
    this->buildGradientResidualAoa(dRu_A);
    // Partials wrt mesh
    dRu_X = Eigen::SparseMatrix<double, Eigen::RowMajor>(sol->pbl->msh->nodes.size(), sol->pbl->nDim * sol->pbl->msh->nodes.size());
    this->buildGradientResidualMesh(dRu_X);
}

/**
 * @brief Solve for the flow residuals given a gradient (wrt to flow variables)
 *
 * solve for dRu: dRu/dU^T * dRu = dU
 */
std::vector<double> Adjoint::solveFlow(std::vector<double> const &dU)
{
    // Create maps
    Eigen::VectorXd dR(sol->pbl->msh->nodes.size());
    Eigen::Map<Eigen::VectorXd> dR_(dR.data(), dR.size());
    Eigen::VectorXd dU_ = Eigen::Map<const Eigen::VectorXd>(dU.data(), dU.size());
    // Solve
    alinsol->compute(dRu_U.transpose(), Eigen::Map<Eigen::VectorXd>(dU_.data(), dU_.size()), dR_);
    return std::vector<double>(dR.data(), dR.data() + dR.size());
}

/**
 * @brief Compute the matrix-vector product of the flow Jacobian with the flow residuals
 *
 * dU = dRu/dU^T * dRu
 * Note: only for multi-disciplinary, not tested by run()
 */
std::vector<double> Adjoint::computeJacVecFlow(std::vector<double> const &dR)
{
    Eigen::VectorXd JdR = dRu_U.transpose() * Eigen::Map<const Eigen::VectorXd>(dR.data(), dR.size());
    return std::vector<double>(JdR.data(), JdR.data() + JdR.size());
}

/**
 * @brief Compute the matrix-vector product of the gradients of the flow residuals (wrt angle of attack) with the flow residuals
 *
 * dA = dRu/dA^T * dRu
 */
double Adjoint::computeJacVecFlowAoa(std::vector<double> const &dR)
{
    return dRu_A.transpose() * Eigen::Map<const Eigen::VectorXd>(dR.data(), dR.size());
}

/**
 * @brief Compute the matrix-vector product of the gradients of the flow residuals (wrt mesh coordinates) with the flow residuals
 *
 * dX = dRu/dX^T * dRu
 */
std::vector<double> Adjoint::computeJacVecFlowMesh(std::vector<double> const &dR)
{
    Eigen::VectorXd dX = dRu_X.transpose() * Eigen::Map<const Eigen::VectorXd>(dR.data(), dR.size());
    return std::vector<double>(dX.data(), dX.data() + dX.size());
}

/**
 * @brief Compute the partial gradients of the loads, on a given body
 *
 * dL_U = dL/dU (u is the vector of flow variables, ie the potential "phi")
 * dL_X = dL/dX (x is the vector of mesh coordinates)
 * Note: only for multi-disciplinary, not tested by run()
 */
void Adjoint::linearizeLoads(Body const &bnd)
{
    // Build gradients wrt flow variables
    dL_U = Eigen::SparseMatrix<double, Eigen::RowMajor>(3 * bnd.nodes.size(), sol->pbl->msh->nodes.size());
    this->buildGradientLoadsFlow(dL_U, bnd);
    // Build gradients wrt to mesh coordinates
    dL_X = Eigen::SparseMatrix<double, Eigen::RowMajor>(3 * bnd.nodes.size(), sol->pbl->nDim * sol->pbl->msh->nodes.size());
    this->buildGradientLoadsMesh(dL_X, bnd);
}

/**
 * @brief Compute the matrix-vector product of the gradients of the loads (wrt flow variables) with the loads
 *
 * dU = dL/dU^T * dU
 * Note: only for multi-disciplinary, not tested by run()
 * Note: only for body passed to linearizeLoads()
 */
std::vector<double> Adjoint::computeJacVecLoadsFlow(std::vector<double> const &dL)
{
    Eigen::VectorXd dU = dL_U.transpose() * Eigen::Map<const Eigen::VectorXd>(dL.data(), dL.size());
    return std::vector<double>(dU.data(), dU.data() + dU.size());
}

/**
 * @brief Compute the matrix-vector product of the gradients of the loads (wrt the mesh coordinates) with the loads
 *
 * dX = dL/dX^T * dL
 * Note: only for multi-disciplinary, not tested by run()
 * Note: only for body passed to linearizeLoads()
 */
std::vector<double> Adjoint::computeJacVecLoadsMesh(std::vector<double> const &dL)
{
    Eigen::VectorXd dX = dL_X.transpose() * Eigen::Map<const Eigen::VectorXd>(dL.data(), dL.size());
    return std::vector<double>(dX.data(), dX.data() + dX.size());
}

/**
 * @brief Compute the gradients of the load coefficients wrt flow variables
 *
 * dCl/dU, dCd/dU
 */
std::vector<std::vector<double>> Adjoint::computeGradientCoefficientsFlow()
{
    // Compute gradients of lift and drag
    Eigen::RowVectorXd dCl = Eigen::RowVectorXd::Zero(sol->pbl->msh->nodes.size()), dCd = Eigen::RowVectorXd::Zero(sol->pbl->msh->nodes.size());
    for (auto bnd : sol->pbl->bodies)
        this->buildGradientCoefficientsFlow(dCl, dCd, *bnd);
    return std::vector<std::vector<double>>{std::vector<double>(dCl.data(), dCl.data() + dCl.size()), std::vector<double>(dCd.data(), dCd.data() + dCd.size())};
}

/**
 * @brief Compute the gradients of the load coefficients wrt angle of attack
 *
 * dCl/dA, dCd/dA
 */
std::vector<double> Adjoint::computeGradientCoefficientsAoa()
{
    double dCl, dCd;
    this->buildGradientLoadsAoa(dCl, dCd);
    return std::vector<double>{dCl, dCd};
}

/**
 * @brief Compute the gradients of the load coefficients wrt mesh coordinates
 *
 * dCl/dX, dCd/dX
 */
std::vector<std::vector<double>> Adjoint::computeGradientCoefficientsMesh()
{
    // Compute the gradients of the lift and drag
    Eigen::RowVectorXd dCl = Eigen::RowVectorXd::Zero(sol->pbl->nDim * sol->pbl->msh->nodes.size()), dCd = Eigen::RowVectorXd::Zero(sol->pbl->nDim * sol->pbl->msh->nodes.size());
    for (auto bnd : sol->pbl->bodies)
        this->buildGradientCoefficientsMesh(dCl, dCd, *bnd);
    return std::vector<std::vector<double>>{std::vector<double>(dCl.data(), dCl.data() + dCl.size()), std::vector<double>(dCd.data(), dCd.data() + dCd.size())};
}

/**
 * @brief Build gradients of the loads with respect to the flow variables on a given body
 */
void Adjoint::buildGradientLoadsFlow(Eigen::SparseMatrix<double, Eigen::RowMajor> &dL, Body const &bnd)
{
    // Multithread
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);
    tbb::spin_mutex mutex;

    // Build gradients
    std::deque<Eigen::Triplet<double>> T;
    tbb::parallel_for_each(bnd.groups[0]->tag->elems.begin(), bnd.groups[0]->tag->elems.end(), [&](Element *e) {
        // Associated volume element
        Element *eV = bnd.svMap.at(e);
        // Build
        Eigen::MatrixXd Ae = LoadFunctional::buildGradientFlow(*e, *eV, sol->phi, *sol->pbl->fluid->cP);
        // Assembly
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            int rowi = static_cast<int>(bnd.nMap.at(e->nodes[i]));
            for (int m = 0; m < 3; ++m)
            {
                for (size_t j = 0; j < eV->nodes.size(); ++j)
                {
                    int rowj = eV->nodes[j]->row;
                    T.push_back(Eigen::Triplet<double>(3 * rowi + m, rowj, Ae(3 * i + m, j)));
                }
            }
        }
    });
    dL.setFromTriplets(T.begin(), T.end());
    dL.prune(0.);
    dL.makeCompressed();
}

/**
 * @brief Build gradients of the load coefficients with respect to the flow variables on a given body
 *
 * Note: this is essentially the same as buildGradientLoadsFlow, but with an additional projection performed inside the loop for performance
 */
void Adjoint::buildGradientCoefficientsFlow(Eigen::RowVectorXd &dCl, Eigen::RowVectorXd &dCd, Body const &bnd)
{
    // Multithread
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);
    tbb::spin_mutex mutex;

    // Build gradients
    Eigen::RowVector3d dirL = sol->pbl->dirL.eval().transpose(), dirD = sol->pbl->dirD.eval().transpose();
    tbb::parallel_for_each(bnd.groups[0]->tag->elems.begin(), bnd.groups[0]->tag->elems.end(), [&](Element *e) {
        // Associated volume element
        Element *eV = bnd.svMap.at(e);
        // Build
        Eigen::MatrixXd Ae = LoadFunctional::buildGradientFlow(*e, *eV, sol->phi, *sol->pbl->fluid->cP);
        // Project
        Eigen::RowVectorXd dCle = Eigen::RowVectorXd::Zero(eV->nodes.size()), dCde = Eigen::RowVectorXd::Zero(eV->nodes.size());
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            dCle += dirL * Ae.middleRows(i * 3, 3);
            dCde += dirD * Ae.middleRows(i * 3, 3);
        }
        // Assembly
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t j = 0; j < eV->nodes.size(); ++j)
        {
            int rowj = eV->nodes[j]->row;
            dCl(rowj) += dCle(j);
            dCd(rowj) += dCde(j);
        }
    });
}

/**
 * @brief Build gradients of the residual with respect to angle of attack
 */
void Adjoint::buildGradientResidualAoa(Eigen::VectorXd &dR)
{
    // Multithread
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);
    tbb::spin_mutex mutex;

    // Reset gradient
    dR.setZero();

    // Freestream contribution
    for (auto nBC : sol->pbl->fBCs)
    {
        tbb::parallel_for_each(nBC->tag->elems.begin(), nBC->tag->elems.end(), [&](Element *e) {
            // Build elementary flux vector
            Eigen::VectorXd be = FreestreamResidual::buildGradientAoa(*e, sol->phi, *nBC);

            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < e->nodes.size(); ++ii)
            {
                Node *nodi = e->nodes[ii];
                dR(sol->rows[nodi->row]) += be(ii);
            }
        });
    }
}

/**
 * @brief Build gradients of the lift and drag with respect to angle of attack
 */
void Adjoint::buildGradientLoadsAoa(double &dCl, double &dCd)
{
    // Compute integrated aerodynamic load coefficients (normalized by freestream dynamic pressure and reference area)
    Eigen::Vector3d Cf(0, 0, 0);
    for (auto bnd : sol->pbl->bodies)
        for (size_t i = 0; i < bnd->nodes.size(); ++i)
            Cf += bnd->nLoads[i];
    // Rotate to gradient of flow direction
    dCl = Cf.dot(sol->pbl->dirL.evalGrad());
    dCd = Cf.dot(sol->pbl->dirD.evalGrad());
}

/**
 * @brief Build gradients of the residual with respect to mesh coordinates
 */
void Adjoint::buildGradientResidualMesh(Eigen::SparseMatrix<double, Eigen::RowMajor> &dR)
{
    // Multithread
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);
    tbb::spin_mutex mutex;

    // List of triplets
    std::deque<Eigen::Triplet<double>> T;

    // Volume
    auto fluid = sol->pbl->fluid;
    tbb::parallel_for_each(fluid->adjMap.begin(), fluid->adjMap.end(), [&](std::pair<Element *, std::vector<Element *>> p) {
        // Current element
        Element *e = p.first;
        //Upwind element
        Element *eU = p.second[0];
        // Build elementary matrices
        Eigen::MatrixXd Ae1, Ae2;
        std::tie(Ae1, Ae2) = PotentialResidual::buildGradientMesh(*e, *eU, sol->phi, *fluid, sol->pbl->nDim, sol->muCv, sol->mCOv);

        // Assembly (subsonic)
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            size_t rowi = e->nodes[i]->row;
            for (size_t j = 0; j < e->nodes.size(); ++j)
            {
                int rowj = e->nodes[j]->row;
                for (int n = 0; n < sol->pbl->nDim; ++n)
                    T.push_back(Eigen::Triplet<double>(sol->rows[rowi], sol->pbl->nDim * rowj + n, Ae1(i, sol->pbl->nDim * j + n)));
            }
        }
        // Assembly (supersonic)
        double mach = fluid->mach->eval(*e, sol->phi, 0);
        if (mach > sol->mCOv)
        {
            for (size_t i = 0; i < e->nodes.size(); ++i)
            {
                size_t rowi = e->nodes[i]->row;
                for (size_t j = 0; j < eU->nodes.size(); ++j)
                {
                    int rowj = eU->nodes[j]->row;
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                        T.push_back(Eigen::Triplet<double>(sol->rows[rowi], sol->pbl->nDim * rowj + n, Ae2(i, sol->pbl->nDim * j + n)));
                }
            }
        }
    });
    // Farfield B.C. are not taken into account because they have no influence on the solution

    // Wake
    for (auto wake : sol->pbl->wBCs)
    {
        tbb::parallel_for_each(wake->wEle.begin(), wake->wEle.end(), [&](WakeElement *we) {
            // Build elementary matrices
            Eigen::MatrixXd Aeu, Ael, Aes;
            std::tie(Aeu, Ael, Aes) = WakeResidual::buildGradientMesh(*we, sol->phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t i = 0; i < we->nRow; ++i)
            {
                int rowi = we->wkN[i].second->row;
                for (size_t j = 0; j < we->nColUp; ++j)
                {
                    int rowj = we->volUpE->nodes[j]->row;
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                        T.push_back(Eigen::Triplet<double>(rowi, sol->pbl->nDim * rowj + n, Aeu(i, sol->pbl->nDim * j + n)));
                }
                for (size_t j = 0; j < we->nColLw; ++j)
                {
                    int rowj = we->volLwE->nodes[j]->row;
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                        T.push_back(Eigen::Triplet<double>(rowi, sol->pbl->nDim * rowj + n, -Ael(i, sol->pbl->nDim * j + n)));
                }
                for (size_t j = 0; j < we->surUpE->nodes.size(); ++j)
                {
                    int rowj = we->surUpE->nodes[j]->row;
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                        T.push_back(Eigen::Triplet<double>(rowi, sol->pbl->nDim * rowj + n, Aes(i, sol->pbl->nDim * j + n)));
                }
            }
        });
    }
    // Kutta
    for (auto kutta : sol->pbl->kSCs)
    {
        tbb::parallel_for_each(kutta->kEle.begin(), kutta->kEle.end(), [&](KuttaElement *ke) {
            // Build elementary matrices
            Eigen::MatrixXd Ae, Aes;
            std::tie(Ae, Aes) = KuttaResidual::buildGradientMesh(*ke, sol->phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t i = 0; i < ke->nRow; ++i)
            {
                int rowi = ke->teN[i].second->row;
                for (size_t j = 0; j < ke->nCol; ++j)
                {
                    int rowj = ke->volE->nodes[j]->row;
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                        T.push_back(Eigen::Triplet<double>(rowi, sol->pbl->nDim * rowj + n, Ae(i, sol->pbl->nDim * j + n)));
                }
                for (size_t j = 0; j < ke->surE->nodes.size(); ++j)
                {
                    int rowj = ke->surE->nodes[j]->row;
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                        T.push_back(Eigen::Triplet<double>(rowi, sol->pbl->nDim * rowj + n, Aes(i, sol->pbl->nDim * j + n)));
                }
            }
        });
    }
    // Assemble matrix
    dR.setFromTriplets(T.begin(), T.end());
    dR.prune(0.);
    dR.makeCompressed();
}

/**
 * @brief Build gradients of the loads with respect to mesh coordinates on a given body
 */
void Adjoint::buildGradientLoadsMesh(Eigen::SparseMatrix<double, Eigen::RowMajor> &dL, Body const &bnd)
{
    // Multithread
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);
    tbb::spin_mutex mutex;

    // Build gradients
    std::deque<Eigen::Triplet<double>> T;
    tbb::parallel_for_each(bnd.groups[0]->tag->elems.begin(), bnd.groups[0]->tag->elems.end(), [&](Element *e) {
        // Volume element
        Element *eV = bnd.svMap.at(e);
        // Build
        Eigen::MatrixXd Aes, Aev;
        std::tie(Aes, Aev) = LoadFunctional::buildGradientMesh(*e, *eV, sol->phi, *sol->pbl->fluid->cP, sol->pbl->nDim);
        // Assembly
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            int rowi = static_cast<int>(bnd.nMap.at(e->nodes[i]));
            for (int m = 0; m < 3; ++m)
            {
                // surface
                for (size_t j = 0; j < e->nodes.size(); ++j)
                {
                    int rowj = e->nodes[j]->row;
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                        T.push_back(Eigen::Triplet<double>(3 * rowi + m, sol->pbl->nDim * rowj + n, Aes(3 * i + m, sol->pbl->nDim * j + n)));
                }
                // volume
                for (size_t j = 0; j < eV->nodes.size(); ++j)
                {
                    int rowj = eV->nodes[j]->row;
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                        T.push_back(Eigen::Triplet<double>(3 * rowi + m, sol->pbl->nDim * rowj + n, Aev(3 * i + m, sol->pbl->nDim * j + n)));
                }
            }
        }
    });
    dL.setFromTriplets(T.begin(), T.end());
    dL.prune(0.);
    dL.makeCompressed();
}

/**
 * @brief Build gradients of the loads with respect to mesh coordinates on a given body
 *
 * Note: this is essentially the same as buildGradientLoadsMesh, but with an additional projection performed inside the loop for performance
 */
void Adjoint::buildGradientCoefficientsMesh(Eigen::RowVectorXd &dCl, Eigen::RowVectorXd &dCd, Body const &bnd)
{
    // Multithread
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);
    tbb::spin_mutex mutex;

    // Build gradients
    Eigen::RowVector3d dirL = sol->pbl->dirL.eval().transpose(), dirD = sol->pbl->dirD.eval().transpose();
    tbb::parallel_for_each(bnd.groups[0]->tag->elems.begin(), bnd.groups[0]->tag->elems.end(), [&](Element *e) {
        // Associated volume element
        Element *eV = bnd.svMap.at(e);
        // Build
        Eigen::MatrixXd Aes, Aev;
        std::tie(Aes, Aev) = LoadFunctional::buildGradientMesh(*e, *eV, sol->phi, *sol->pbl->fluid->cP, sol->pbl->nDim);
        // Project
        Eigen::RowVectorXd dCles = Eigen::RowVectorXd::Zero(sol->pbl->nDim * e->nodes.size()), dCdes = Eigen::RowVectorXd::Zero(sol->pbl->nDim * e->nodes.size());
        Eigen::RowVectorXd dClev = Eigen::RowVectorXd::Zero(sol->pbl->nDim * eV->nodes.size()), dCdev = Eigen::RowVectorXd::Zero(sol->pbl->nDim * eV->nodes.size());
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            dCles += dirL * Aes.middleRows(i * 3, 3);
            dCdes += dirD * Aes.middleRows(i * 3, 3);
            dClev += dirL * Aev.middleRows(i * 3, 3);
            dCdev += dirD * Aev.middleRows(i * 3, 3);
        }
        // Assembly
        tbb::spin_mutex::scoped_lock lock(mutex);
        // surface
        for (size_t j = 0; j < e->nodes.size(); ++j)
        {
            int rowj = e->nodes[j]->row;
            for (int n = 0; n < sol->pbl->nDim; ++n)
            {
                dCl(sol->pbl->nDim * rowj + n) += dCles(sol->pbl->nDim * j + n);
                dCd(sol->pbl->nDim * rowj + n) += dCdes(sol->pbl->nDim * j + n);
            }
        }
        // volume
        for (size_t j = 0; j < eV->nodes.size(); ++j)
        {
            int rowj = eV->nodes[j]->row;
            for (int n = 0; n < sol->pbl->nDim; ++n)
            {
                dCl(sol->pbl->nDim * rowj + n) += dClev(sol->pbl->nDim * j + n);
                dCd(sol->pbl->nDim * rowj + n) += dCdev(sol->pbl->nDim * j + n);
            }
        }
    });
}

/**
 * @brief Write the results
 */
void Adjoint::save(MshExport *mshWriter, std::string const &suffix)
{
    // Write files
    std::cout << "Saving files "
              << std::setprecision(2)
              << "(Mach " << sol->pbl->M_inf << ", "
              << sol->pbl->alpha * 180 / 3.14159 << " deg AoA, "
              << sol->pbl->beta * 180 / 3.14159 << " deg AoS)"
              << std::endl;
    // setup results
    Results results;
    results.scalars_at_nodes["lambdaClPhi"] = &lamClFlo;
    results.scalars_at_nodes["lambdaCdPhi"] = &lamCdFlo;
    results.vectors_at_nodes["gradClMsh"] = &dClMsh;
    results.vectors_at_nodes["gradCdMsh"] = &dCdMsh;
    // save (whole mesh and bodies)
    mshWriter->save(sol->pbl->msh->name + "_adjoint" + suffix, results);
    for (auto bnd : sol->pbl->bodies)
        bnd->save(sol->pbl->msh->name + '_' + bnd->groups[0]->tag->name + "_adjoint" + suffix, results);
}

void Adjoint::write(std::ostream &out) const
{
    out << "dart::Adjoint"
        << "\n";
}
