#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Initialize VII computation
# Paul Dechamps

def initVII(cfg, icfg, iSolverName='DART'):
    """
    Inputs
    ------
    - cfg: Dict with options
        - Re (float): Flow Reynolds number.
        - Minf (float): Freestream Mach number (only used for time step computation).
        - CFL0 (float): Initial CFL number.
        - Verb (int): Verbosity level of the viscous solver.

        - couplIter (int): Maximum number of coupling iterations.
        - couplTol (float): Tolerance to end computation (drag count between 2 iterations).
        - iterPrint (int): Number of iterations between which the coupler outputs its state.
        - resetInv (bool): Resets the inviscid solver at each iteration (True), reuses previous
                            state as initial condition (False).
        Note
        All inputs have default values except Re.
    - icg: Dict with options. Inviscid solver configuration.
    - iSolverName (string): Name of the inviscid solver.

    Outputs
    -------
    dict with keys
        - coupler: Coupler python object to run the viscous-inviscid algorithm.
        - solversAPI: Interface between the solver objects to ensure proper communication.
        - vSolver: vii::Driver object to run the viscous computation.
        - iSolverObjects: Dict containing.
            - All inviscid solver dependencies (depend on iSolverName).

    @todo: - Correctly handle 3D computation parameters.
    """
    # Imports
    from vii.coupler import Coupler
    import vii

    if iSolverName == 'DART':
        from vii.interfaces.dart.DartInterface import DartInterface as interface
        from dart.api.core import initDart
        iSolverObjects = initDart(icfg, scenario=icfg['scenario'], task=icfg['task'], viscous = 1)

    # Check viscous solver parameters.
    if 'Re' in cfg and cfg['Re'] > 0:
        _Re = cfg['Re']
    else:
        raise RuntimeError('Missing or invalid Reynolds number')
    if 'Minf' in cfg and cfg['Minf'] > 0:
        _Minf = cfg['Minf']
    else:
        _Minf = 0.1
    if 'CFL0' in cfg and cfg['CFL0'] > 0:
        _CFL0 = cfg['CFL0']
    else:
        _CFL0 = 1
    if 'Verb' in cfg and 0<= cfg['Verb'] <= 3:
        _verbose = cfg['Verb']
    else:
        _verbose = 1
    _span = 0
    _nSections = 1 

    # Check coupler parameters.
    if 'couplIter' in cfg and cfg['couplIter'] > 0:
        __couplIter = cfg['couplIter']
    else:
        __couplIter = 150
    if 'couplTol' in cfg:
        __couplTol = cfg['couplTol']
    else:
        __couplTol = 1e-4
    if 'iterPrint' in cfg:
        __iterPrint = cfg['iterPrint']
    else:
        __iterPrint = 1
    if 'resetInv' in cfg:
        __resetInv = cfg['resetInv']
    else:
        __resetInv = False

    # Viscous solver object.
    vSolver = vii.Driver(_Re, _Minf, _CFL0, _nSections, _span, verbose =_verbose)
    # Solvers interface.
    solversAPI = interface(iSolverObjects['sol'], vSolver, [iSolverObjects['blwb'], iSolverObjects['blww']])
    # Coupler
    coupler = Coupler(solversAPI, vSolver, _maxCouplIter = __couplIter, _couplTol = __couplTol, _iterPrint = __iterPrint, _resetInv = __resetInv)

    return {'coupler': coupler,
            'solversAPI': solversAPI,
            'vSolver': vSolver,
            'iSolverObjects': iSolverObjects}
